#!/usr/bin/env ruby

require 'rdparser'
require 'tabtest'

##############################################################################
#
# This part defines the TabTab gates
#
##############################################################################

@@stop = ["**return**","**break**","**skip**"]
@@type_convert = {
  Fixnum=>"int",
  Float=>"float",
  String=>"string",
  Array=>"array",
  TrueClass=>"bool",
  FalseClass=>"bool"}

def check_scope(name)
  value = [false,nil]
  index = @@scope.size
  while index > 0
    index -= 1
    if @@scope[index].has_key?(name)
      value = [true,index]
      return value
    end
  end
  return value
end


def array_check(type,value)
  right = true
  for element in value
    if @@type_convert[element.class] != type
      right = false
    end
  end
  return right
end


def type_check(name,value,type,is_array)
  set = false
  value_type = @@type_convert[value.class]
  if is_array
    if value_type == "array"
      @@scope[@@current_scope][name] = ["array", value, type]
      set = true
    end
  else
    if value_type == type
      @@scope[@@current_scope][name] = [type, value]
      set = true
    end
  end

  if not set
    raise "TTParser: Variable '#{name}' of type '#{type}' cannot be set to value '#{value}'."
  end
end


class Program

  def initialize(stmts)
    @stmts = stmts
    eval()
  end

  def eval
    @stmts.eval()
  end
end


class Statements

  def initialize(stmt,stmts=nil)
    @stmt = stmt
    @stmts = stmts
  end

  def eval
    if(@stmts != nil)
      var = @stmts.eval()
      if var.class == Array
        if @@stop.include?(var[0])
          return var
        end
      end
    end
    var = @stmt.eval()
    if var.class == Array
      if @@stop.include?(var[0])
        return var
      end
    end
  end
end


class Statement

  def initialize(stmt)
    @stmt = stmt
  end

  def eval
    var = @stmt.eval()
    if var.class == Array
      if @@stop.include?(var[0])
        return var
      end
    end
  end
end


class Return

  def initialize(value)
    @value = value
  end

  def eval
    return ["**return**",@value]
  end
end


class Break

  def initialize
  end

  def eval
    return ["**break**"]
  end
end


class Skip

  def initialize
  end

  def eval
    return ["**skip**"]
  end
end


class Comment

  def initialize(comment)
    @comment = comment
  end

  def eval
    #Debug info?
    nil
  end
end


class Def_fun

  def initialize(type, name, stmts, paras=nil, is_array=false)
    @type = type
    @name = name
    @stmts = stmts
    @paras = paras
    @is_array = is_array
  end

  def eval
    if @is_array
      @@scope[@@current_scope][@name.eval()] = [Type.new("array"), @paras, @stmts, @type]
    else
      @@scope[@@current_scope][@name.eval()] = [@type, @paras, @stmts]
    end
  end
end


class Call_fun

  def initialize(name, args=nil)
    @name = name
    @args = args
  end

  def eval
    if check_scope(@name.eval())[0]
      has_returnval = false
      index = check_scope(@name.eval())[1]
      @type = @@scope[index][@name.eval()][0]
      stmts = @@scope[index][@name.eval()][2] #Retrieve function statements
      @@scope << {}
      @@current_scope += 1
      if @args != nil
        #Create variables if arguments are included
        arguments = @args.eval()
        parameters = @@scope[index][@name.eval()][1].eval()

        for index in (0...arguments.size)
          value = arguments[index]
          name = parameters[index][1]
          type = parameters[index][0]
          if @@type_convert[value.class] == "array"
            arr_type = parameters[2]
            type_check(name,value,arr_type,true)
          else
            type_check(name,value,type,false)
          end
        end
      end
      var = stmts.eval() #Run function statements
      if var.class == Array
        if var[0] == "**return**"
          has_returnval = true
        end
        if @@stop.include?(var[0]) and var[0] != "**return**"
          return var
        end
      end
      if has_returnval
        index = check_scope(@name.eval())[1]
        rtrn_value = var[1].eval()
        if @@type_convert[rtrn_value.class] == @type.eval()
          if @type.eval() == "array" and array_check(@@scope[index][@name.eval()][3].eval(), rtrn_value)
            return rtrn_value
          else
            return rtrn_value
          end
        else
          raise "TTParser: Expected return value of type '#{@type.eval()}' from function '#{@name.eval()}', found '#{@@type_convert[rtrn_value.class]}'!"
        end
      else
        if @type.eval() != "void"
          raise "TTParser: Expected return value of type '#{@type.eval()}' from function '#{@name.eval()}', found none!"
        else
          return nil
        end
      end
      @@current_scope -= 1
      @@scope.pop()
      nil
    else
      raise "TTParser: Invalid function call. Function '#{@name.eval()}' does not exist!"
    end
  end
end


class Parameters

  def initialize(para, paras=nil)
    @para = para
    @paras = paras
  end

  def eval
    if @paras != nil
      return @paras.eval() + [@para.eval()]
    else
      return [@para.eval()]
    end
  end
end


class Parameter

  def initialize(type, name, is_array=false)
    @type = type
    @name = name
    @is_array = is_array
  end

  def eval
    if @is_array
      return ["array", @name.eval(), @type.eval()]
    else
      return [@type.eval(),@name.eval()]
    end
  end
end


class Arguments

  def initialize(arg, args=nil)
    @arg = arg
    @args = args
  end

  def eval
    if @args != nil
      return @args.eval() + @arg.eval()
    else
      return @arg.eval()
    end
  end
end


class Argument

  def initialize(value)
    @value = value
  end

  def eval
     return [@value.eval()]
  end
end


class Output

  def initialize(exprs)
    @exprs = exprs
  end

  def eval
    for expr in @exprs.eval()
      if expr.eval() != nil
        puts expr.eval()
      end
    end
  end
end


class Input

  def initialize(var,typed=false)
    @var = var
    @typed = typed
  end

  def eval
    if @typed
      value = gets
      @@scope[@@current_scope][@var.eval()] = ["string", value[0...-1]]
    else
      if check_scope(@var.eval())[0]
        value = gets
        index = check_scope(@var.eval())[1]
        type = @@scope[index][@var.eval()][0]
        if type == "string"
          @@scope[index][@var.eval()][1] = value[0...-1]
        else
          raise "TTParser: Variable '#{@var.eval()}' is not of type string."
        end
      else
        raise "TTParser: Variable '#{@var.eval()}' is not defined."
      end
    end
  end
end


class For_stmt

  def initialize(name,container,stmts)
    @name = name
    @container = container
    @stmts = stmts
  end

  def get_element(cont,i)
    if cont.class == Array
      return cont[i..i][0]
    else
      return cont[i..i]
    end
  end

  def eval
    @@scope << {}
    @@current_scope += 1
    cont = @container.eval()
    @type = @@type_convert[cont[0].class]
    for i in (0...cont.size)
      @@scope[@@current_scope][@name.eval()] = [@type, get_element(cont,i)]
      var = @stmts.eval()
      if var.class == Array
        if var[0] == "**break**"
          break
        end
        if var[0] == "**skip**"
          next
        end
        if @@stop.include?(var[0]) and var[0] != "**break**" and var[0] != "**skip**"
          return var
        end
      end
    end
    @@current_scope -= 1
    @@scope.pop()
    nil
  end
end


class While_stmt

  def initialize(expr,stmts)
    @expr = expr
    @stmts = stmts
  end

  def eval
    @@scope << {}
    @@current_scope += 1
    while @expr.eval() do
      var = @stmts.eval()
      if var.class == Array
        if var[0] == "**break**"
          break
        end
        if var[0] == "**skip**"
          next
        end
        if @@stop.include?(var[0]) and var[0] != "**break**" and var[0] != "**skip**"
          return var
        end
      end
    end
    @@current_scope -= 1
    @@scope.pop()
    nil
  end
end


class If_stmt

  def initialize(expr,stmts,stmts_else=nil)
    @expr = expr
    @stmts = stmts
    @stmts_else = stmts_else
  end

  def eval
    if @stmts_else == nil
      if @expr.eval()
        @stmts.eval()
      end
    else
      if @expr.eval()
        @stmts.eval()
      else
        @stmts_else.eval()
      end
    end
  end
end


class Def_var

  def initialize(name, value, type=nil, is_array=false)
    @type = type
    @name = name
    @value = value
    @is_array = is_array
  end

  def eval
    if @type != nil
      if @is_array
        if @value.check_types(@type.eval())
          type_check(@name.eval(),@value.eval(),@type.eval(),@is_array)
        end
      else
        if @@types.include?(@value.eval().class)
          type_check(@name.eval(),@value.eval(),@type.eval(),@is_array)
        end
      end
    else
      if check_scope(@name.eval())[0]
        index = check_scope(@name.eval())[1]
        value = @value.eval()
        type = @@scope[index][@name.eval()][0]
        if type == @@type_convert[value.class]
          if type == "array"
            array_type = @@scope[index][@name.eval()][2]
            if array_check(array_type, value)
              @@scope[index][@name.eval()][1] = value
            else
              raise "TTParser: Element(s) in '#{value}' is not of type '#{array_type}'."
            end
          else
            @@scope[index][@name.eval()][1] = value
          end
        else
          raise "TTParser: Variable '#{@name.eval()}' of type '#{type}' cannot be set to value '#{value}'."
        end
      else
        raise "TTParser: Variable '#{@name.eval()}' is not defined."
      end
    end
  end
end


class Output_exprs

  def initialize(expr,exprs=nil)
    @expr = expr
    @exprs = exprs
  end

  def eval
    if @exprs != nil
      return @exprs.eval() + [@expr]
    end
    return [@expr]
  end
end


class Expression

  def initialize(expr)
    @expr = expr
  end

  def eval
    @expr.eval()
  end
end


class Comp_expr

  def initialize(expr,op=nil,expr2=nil)
    @expr = expr
    @op = op
    @expr2 = expr2
  end

  def eval
    if @op == nil or @expr2 == nil
      @expr.eval()
    else
      operator = @op.eval()
      case operator
        when "<" then @expr.eval() < @expr2.eval()
        when ">" then @expr.eval() > @expr2.eval()
        when "==" then @expr.eval() == @expr2.eval()
        when "<=" then @expr.eval() <= @expr2.eval()
        when ">=" then @expr.eval() >= @expr2.eval()
        when "!=" then @expr.eval() != @expr2.eval()
      end
    end
  end
end


class Comp_operator

  def initialize(value)
    @value = value
  end

  def eval
    @value
  end
end


class Calc_expr

  def initialize(expr,op=nil,expr2=nil)
    @expr = expr
    @op = op
    @expr2 = expr2
  end

  def eval
    if @op == nil or @expr2 == nil
      @expr.eval()
    else
      operator = @op
      case operator
        when "+" then @expr.eval() + @expr2.eval()
        when "-" then @expr.eval() - @expr2.eval()
        when "*" then @expr.eval() * @expr2.eval()
        when "/" then @expr.eval() / @expr2.eval()
      end
    end
  end
end


class Neg_expr

  def initialize(expr,op=nil)
    @expr = expr
    @op = op
  end

  def eval
    if @op == "-"
      return -(@expr.eval())
    else
      return @expr.eval()
    end
  end
end


class Log_expr

  def initialize(expr,op=nil,expr2=nil)
    @expr = expr
    @op = op
    @expr2 = expr2
  end

  def eval
    if @op == nil or @expr2 == nil
      @expr.eval()
    else
      operator = @op
      case operator
        when "and" then @expr.eval() and @expr2.eval()
        when "or" then @expr.eval() or @expr2.eval()
      end
    end
  end
end


class Not_expr

  def initialize(expr,op=nil)
    @expr = expr
    @op = op
  end

  def eval
    if @op == "not"
      return (not @expr.eval())
    else
      return @expr.eval()
    end
  end
end


class Type

  def initialize(value)
    @value = value
  end

  def eval
    @value
  end
end


class Int_type

  def initialize(value)
    @value = value
  end

  def eval
    @value
  end
end


class Float_type

  def initialize(value)
    @value = value
  end

  def eval
    @value
  end
end


class String_type

  def initialize(value)
    @value = value
  end

  def eval
    @value.slice(1,@value.size-2)
  end
end


class Bool_type

  def initialize(value)
    @value = value
  end

  def eval
    @value
  end
end


class Array_elems

  def initialize(arg, args=nil)
    @arg = arg
    @args = args
  end

  def eval
    if @args != nil
      return @args.eval() + @arg.eval()
    else
      return @arg.eval()
    end
  end
end


class Array_elem

  def initialize(value)
    @value = value
  end

  def eval
      return [@value.eval()]
  end
end


class Array_type

  def initialize(elements=nil)
    @elements = elements
  end

  def eval
    if @elements != nil
      return @elements.eval()
    else
      return []
    end
  end

  def check_types(type)
    right = true
    for element in @elements.eval()
      if @@type_convert[element.class] != type
        right = false
        raise "TTParser: Element '#{element}' is of type '#{@@type_convert[element.class]}', expected '#{type}'!"
      end
    end
    return right
  end
end


class Identifier

  def initialize(value)
    @value = value
  end

  def eval
    @value
  end
end


class Variable

  def initialize(value)
    @value = value
  end

  def eval
    var = @value.eval()
    if check_scope(var)[0]
      return @@scope[check_scope(var)[1]][var][1]
    else
      raise "TTParser: No variable named '#{var}'."
    end
  end
end


class Container

  def initialize(cont)
    @cont = cont
  end

  def eval
    value = @cont.eval()
    type = @@type_convert[value.class]
    if type == "array" or type == "string"
      return value
    else
      raise "TTParser: Variable is not a valid container."
    end
  end
end



##############################################################################
#
# This part defines the TabTab language
#
##############################################################################
@@scope = []
@@scope << {}
@@current_scope = 0
@@types = [Fixnum,Float,String,Array,TrueClass,FalseClass]

class TabTab

  def initialize
    @TTParser = Parser.new("TabTab") do
      #Newline
      token(/\s*\n+/) {"\n"}
      #Comment
      token(/\/\/.*\s*\n+/) {|m| m}
      #Remove whitespace
      token(/\s+/)
      token(/ /)
      #Float
      token(/\d+\.\d+/) {|m| m.to_f}
      #Int
      token(/\d+/) {|m| m.to_i}
      #String
      token(/"[^"]*[^\\]"/) {|m| m} #crash on "\\"
      token(/'[^']*[^\\]'/) {|m| m}
      #Parenthesis
      token(/[\(\)]/) {|m| m}
      #Calc Operators
      token(/\+/) {|m| m}
      token(/-/) {|m| m}
      token(/\*/) {|m| m}
      token(/\//) {|m| m}
      #Comp Operators
      token(/<=/) {|m| m}
      token(/>=/) {|m| m}
      token(/==/) {|m| m}
      token(/!=/) {|m| m}
      token(/</) {|m| m}
      token(/>/) {|m| m}
      #Comma
      token(/,/) {|m| m}
      #Assignment Operator
      token(/=/) {|m| m}
      #Array
      token(/\[/) {|m| m}
      token(/\]/) {|m| m}
      #Word
      token(/[a-zA-Z_][a-zA-Z0-9_]*/) {|m| m}
      #Everything else
      token(/.+/) {|m| m}

      start :program do
        match("program", "\n", :statements, "endprogram", "\n"){|_,_,stmts,_,_| Program.new(stmts)
        nil}
      end

      rule :def_fun do
        match("fun", :fun_type, :identifier, "(", :parameters, ")", "\n", :statements, "endfun") {|_,type,name,_,args,_,_,stmts,_| Def_fun.new(type,name,stmts,args)}
        match("fun", :fun_type, :identifier, "(", ")", "\n", :statements, "endfun") {|_,type,name,_,_,_,stmts,_| Def_fun.new(type,name,stmts)}
        match("fun", :array_type, :identifier, "(", :parameters, ")", "\n", :statements, "endfun") {|_,type,name,_,args,_,_,stmts,_| Def_fun.new(type,name,stmts,args,true)}
        match("fun", :array_type, :identifier, "(", ")", "\n", :statements, "endfun") {|_,type,name,_,_,_,stmts,_| Def_fun.new(type,name,stmts,nil,true)}
      end

      rule :call_fun do
        match(:identifier, "(",  :arguments, ")") {|name,_,args,_| Call_fun.new(name,args)}
        match(:identifier, "(", ")") {|name,_,_| Call_fun.new(name)}
      end

      rule :for_stmt do
        match("for", "(",:identifier, "in", :container, ")", "\n", :statements, "endfor") {|_,_,name,_,container,_,_,stmts,_| For_stmt.new(name,container,stmts)}
      end

      rule :while_stmt do
        match("while", "(", :expression, ")", "\n", :statements, "endwhile") {|_,_,expr,_,_,stmt,_| While_stmt.new(expr,stmt)}
      end

      rule :if_stmt do
        match("if", "(", :expression, ")", "\n", :statements, "endif") {|_,_,expr,_,_,stmt,_| If_stmt.new(expr,stmt)}
        match("if", "(", :expression, ")", "\n", :statements, "else", "\n", :statements, "endif") {|_,_,expr,_,_,stmt,_,_,stmt2,_| If_stmt.new(expr,stmt,stmt2)}
      end

      rule :comment do
        match(/\/\/.*\n/ ) {|value| Comment.new(value)}
      end

      rule :def_var do
        match(:type, :identifier, "=", :call_fun) {|type,name,_,value| Def_var.new(name,value,type)}
        match(:identifier, "=", :call_fun) {|name,_,value| Def_var.new(name,value)}
        match(:type, :identifier, "=", :expression) {|type,name,_,value| Def_var.new(name,value,type)}
        match(:identifier, "=", :expression) {|name,_,value| Def_var.new(name,value)}
        match(:type, :identifier, "=", :string) {|type,name,_,value| Def_var.new(name,value,type)}
        match(:identifier, "=", :string) {|name,_,value| Def_var.new(name,value)}
        match(:array_type, :identifier, "=", :array) {|type,name,_,value| Def_var.new(name,value,type,true)}
        match(:identifier, "=", :array) {|name,_,value| Def_var.new(name,value)}#(name,value,nil,array)?
      end

      rule :container do
        match(:string) {|cont| Container.new(cont)}
        match(:array) {|cont| Container.new(cont)}
        match(:variable) {|cont| Container.new(cont)}
      end

      rule :int do
        match(Integer) {|value| Int_type.new(value)}
      end

      rule :float do
        match(Float) {|value| Float_type.new(value)}
      end

      rule :array_elem do
        match(:string) {|value| Array_elem.new(value)}
        match(:expression) {|value| Array_elem.new(value)}
      end

      rule :array_elems do
        match(:array_elems, ",", :array_elem) {|values,_,value| Array_elems.new(value, values)}
        match(:array_elem) {|value| Array_elems.new(value)}
      end

      rule :array do
        match("[", :array_elems, "]") {|_,elements,_| Array_type.new(elements)}
        match("[","]") {Array_type.new()}
      end

      rule :string do
        match(/".*"/) {|value| String_type.new(value)}
        match(/'.*'/) {|value| String_type.new(value)}
      end

      rule :bool do
        match(:true) {|value| Bool_type.new(value)}
        match(:false) {|value| Bool_type.new(value)}
      end

      rule :true do
        match("true") {true}
      end

      rule :false do
        match("false") {false}
      end

      rule :parameter do
        match(:type, :identifier) {|type,name| Parameter.new(type,name)}
        match(:array_type, :identifier) {|type,name| Parameter.new(type,name,true)}
      end

      rule :parameters do
        match(:parameters, ",", :parameter) {|paras,_,para| Parameters.new(para, paras)}
        match(:parameter) {|para| Parameters.new(para)}
      end

      rule :argument do
        match(:variable) {|value| Argument.new(value)}
        match(:expression) {|value| Argument.new(value)}
        match(:string) {|value| Argument.new(value)}
        match(:array) {|value| Argument.new(value)}
      end

      rule :arguments do
        match(:arguments, ",", :argument) {|args,_,arg| Arguments.new(arg, args)}
        match(:argument) {|arg| Arguments.new(arg)}
      end

      rule :expression do
        match(:or_expr) {|expr| Expression.new(expr)}
      end

      rule :para do
        match("(",:add_calc_expr,")") {|_,expr,_| Calc_expr.new(expr)}
      end

      rule :num do
        match(:para) {|expr| Calc_expr.new(expr)}
        match(:int) {|expr| Calc_expr.new(expr)}
        match(:float) {|expr| Calc_expr.new(expr)}
        match(:variable) {|expr| Calc_expr.new(expr)}
      end

      rule :neg_expr do
        match(:num) {|expr| Neg_expr.new(expr)}
        match("-", :num) {|op,expr| Neg_expr.new(expr,op)}
      end

      rule :multi_calc_expr do
        match(:neg_expr) {|expr| Calc_expr.new(expr)}
        match(:multi_calc_expr, "*", :multi_calc_expr) {|expr,op,expr2| Calc_expr.new(expr,op,expr2)}
        match(:multi_calc_expr, "/", :multi_calc_expr) {|expr,op,expr2| Calc_expr.new(expr,op,expr2)}
      end

      rule :add_calc_expr do
        match(:multi_calc_expr)  {|expr| Calc_expr.new(expr)}
        match(:add_calc_expr, "+", :multi_calc_expr) {|expr,op,expr2| Calc_expr.new(expr,op,expr2)}
        match(:add_calc_expr, "-", :multi_calc_expr) {|expr,op,expr2| Calc_expr.new(expr,op,expr2)}
      end

      rule :calc_expr do
        match(:add_calc_expr)
      end

      rule :comp_expr do
        match(:calc_expr, :comp_operator, :calc_expr) {|expr,op,expr2| Comp_expr.new(expr,op,expr2)}
        match(:calc_expr) {|expr| Comp_expr.new(expr)}
        match(:bool) {|expr| Comp_expr.new(expr)}
        match(:variable) {|expr| Comp_expr.new(expr)}
      end

      rule :comp_operator do
        match("<") {|op| Comp_operator.new(op)}
        match(">") {|op| Comp_operator.new(op)}
        match("==") {|op| Comp_operator.new(op)}
        match("!=") {|op| Comp_operator.new(op)}
        match("<=") {|op| Comp_operator.new(op)}
        match(">=") {|op| Comp_operator.new(op)}
      end

      rule :not_expr do
        match(:comp_expr) {|expr| Not_expr.new(expr)}
        match("not", :not_expr) {|op,expr| Not_expr.new(expr,op)}
      end

      rule :and_expr do
        match(:not_expr) {|expr| Log_expr.new(expr)}
        match(:and_expr,"and",:not_expr) {|expr,op,expr2| Log_expr.new(expr,op,expr2)}
      end

      rule :or_expr do
        match(:and_expr) {|expr| Log_expr.new(expr)}
        match(:or_expr,"or",:and_expr) {|expr,op,expr2| Log_expr.new(expr,op,expr2)}
      end

      rule :return do
        match("return", :expression) {|_,expr| Return.new(expr)}
        match("return", :string) {|_,str| Return.new(str)}
        match("return", :call_fun) {|_,call| Return.new(call)}
        match("return", :array) {|_,arr| Return.new(arr)}
      end

      rule :break do
        match("break") {Break.new()}
      end

      rule :skip do
        match("skip") {Skip.new()}
      end

      rule :input do
        match("string", :identifier, "=", "input") {|_,var,_,_| Input.new(var,true)}
        match(:identifier, "=", "input") {|var,_,_| Input.new(var)}
      end

      rule :statements do
        match(:statements, :lined_stmt) {|stmts,stmt| Statements.new(stmt,stmts)}
        match(:lined_stmt)
      end

      rule :lined_stmt do
        match(:statement, "\n")
        match(:statement, :comment)
        match(:comment)
      end

      rule :statement do
        match(:def_fun) {|stmt| Statement.new(stmt)}
        match(:call_fun) {|stmt| Statement.new(stmt)}
        match(:output) {|stmt| Statement.new(stmt)}
        match(:for_stmt) {|stmt| Statement.new(stmt)}
        match(:if_stmt) {|stmt| Statement.new(stmt)}
        match(:while_stmt) {|stmt| Statement.new(stmt)}
        match(:def_var) {|stmt| Statement.new(stmt)}
        match(:return) {|stmt| Statement.new(stmt)}
        match(:break) {|stmt| Statement.new(stmt)}
        match(:skip) {|stmt| Statement.new(stmt)}
        match(:input) {|stmt| Statement.new(stmt)}
        match(:expression) {|stmt| Statement.new(stmt)}
      end

      rule :identifier do
        match(/^(?=.*?[a-zA-Z_][a-zA-Z0-9_]*)((?!^program$|^endprogram$|^fun$|^endfun$|^if$|^else$|^endif$|^while$|^endwhile$|^for$|^in$|^endfor$|^true$|^false$|^void$|^and$|^or$|^not$|^return$|^break$|^skip$|^input$|^int$|^float$|^array$|^string$|^bool$|^output$|"|'|^\/\/.*\n+).)*$/) {|value| Identifier.new(value)}
      end

      rule :variable do
        match(:identifier) {|value| Variable.new(value)}
      end

      rule :fun_type do
        match(:type)
        match("void") {|type| Type.new(type)}
      end

      rule :array_type do
        match("int","array") {|type,_| Type.new(type)}
        match("float","array") {|type,_| Type.new(type)}
        match("string","array") {|type,_| Type.new(type)}
        match("bool","array") {|type,_| Type.new(type)}
      end

      rule :type do
        match("int") {|type| Type.new(type)}
        match("float") {|type| Type.new(type)}
        match("string") {|type| Type.new(type)}
        match("bool") {|type| Type.new(type)}
      end


      rule :output do
        match("output", :output_exprs) {|_,expr| Output.new(expr)}
      end

      rule :output_exprs do
        match(:output_exprs, ",", :expression) {|exprs,_,expr| Output_exprs.new(expr,exprs)}
        match(:output_exprs, ",", :string) {|exprs,_,str| Output_exprs.new(str,exprs)}
        match(:output_exprs, ",", :call_fun) {|exprs,_,call| Output_exprs.new(call,exprs)}
        match(:output_exprs, ",", :array) {|exprs,_,arr| Output_exprs.new(arr,exprs)}
        match(:call_fun) {|call| Output_exprs.new(call)}
        match(:expression) {|expr| Output_exprs.new(expr)}
        match(:string) {|str| Output_exprs.new(str)}
        match(:array) {|arr| Output_exprs.new(arr)}
      end

    end
  end


  def run(new_file=nil)
    if(new_file.split(".")[-1] == "tt")
      if(tab_test(new_file))
        file = File.new(new_file, "r")
        str = ""
        while(line = file.gets)
          str += line
          if((/endprogram *\n*/ =~ line) != nil)
            break
          end
        end
        file.close
        puts "#{@TTParser.parse str}"
      end
    else
      raise "TTParser: Invalid file type, must be of type '.tt'!"
    end
  end

  def log(state = false)
    if state
      @TTParser.logger.level = Logger::DEBUG
    else
      @TTParser.logger.level = Logger::WARN
    end
  end
end
